#if !defined(_ACCURATE_TIME_H_)
#define _ACCURATE_TIME_H_


/*****************************************************************************
		       Copyright (c) John F. Croix, 1997

			      All Rights Reserved

  Permission to use, copy, modify, and distribute this software and its
  documentation for any non-commercial or internal-use-only purpose and without
  fee is hereby granted, provided that the above copyright notice appears in
  all copies and that both the copyright notice and this permission notice
  appear in supporting documentation, and all changes or alterations to the
  software and documentation are clearly documented.  Resale of any part of
  this software or documentation is prohibited without a license.  Licenses may
  be obtained by contacting the author, John F. Croix, at John.Croix@amd.com or
  journeyman@mail.utexas.edu.

  THE AUTHOR, JOHN F. CROIX, DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
  SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR
  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
  DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
  TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OF PERFORMANCE
  OF THIS SOFTWARE.
 *****************************************************************************/



#include <iostream.h>
#include <memory.h>
#include <sys/time.h>
#include <unistd.h>


/*****************************************************************************
 * Structures to help figure out timeouts.  The Expect code uses the time()
 * function, which returns time in seconds.  I'm trying to give a little more
 * resolution to the problem.  Although I can't justify it empirically, my
 * reasoning follows.
 *
 * Basically, we're reading a single character from the input stream, appending
 * the character to the end of the existing text, and trying for a pattern
 * match.  Getting a single character could be a very rapid operation while
 * the rest of the pattern match may not be, especially if the pattern match
 * is in the scripting language and not compiled into C/C++.  Using time() to
 * wrap just the part in which a single character was read could potentially
 * cause gross differences between the actual amount of time used to read the
 * input and the perceived amount.  For example, if the seconds changed between
 * each start invocation of the character reading code and the end of it,
 * each character read will appear to have taken 1 second.  For a 10 second
 * timeout, we'd only get 10 characters read.  Admittedly, this is a
 * pathological case, but that's the way I think.  This way only the
 * microseconds used to read the character will actually be used.
 *
 * Of course, this is only important if we are reading one character at a time.
 *****************************************************************************/


class AccurateTime
  {
    timeval Time;

    AccurateTime AdjustUnits()
      {
	/*
	 * Make the sign of the two values in the time structure match.
	 */
	
	while (((Time.tv_usec < 0) && (Time.tv_sec > 0)) ||
	       ((Time.tv_usec > 0) && (Time.tv_sec < 0)))
	  if (Time.tv_sec > 0) 
	      {
		Time.tv_sec--;
		Time.tv_usec += 1000000;
	      }
	    else
	      {
		Time.tv_sec++;
		Time.tv_usec -= 1000000;
	      }
	return( *this );
      }

  public:
    AccurateTime()
      {
	memset( &Time, sizeof( Time ), 0 );
      }
    AccurateTime(timeval NewTime)
      {
	Time = NewTime;
	AdjustUnits();
      }
    AccurateTime(int Seconds, long Microseconds=0)
      {
	Time.tv_sec = Seconds;
	Time.tv_usec = Microseconds;
      }

    AccurateTime operator-(const AccurateTime &Sub) const
      {
	timeval T = Time;

	T.tv_usec -= Sub.Time.tv_usec;
	T.tv_sec -= Sub.Time.tv_sec;
	return( AccurateTime( T ) );
      }
    AccurateTime operator-=(const AccurateTime &Sub)
      {
	Time.tv_usec -= Sub.Time.tv_usec;
	Time.tv_sec -= Sub.Time.tv_sec;
	return( AdjustUnits() );
      }
    AccurateTime operator+(const AccurateTime &Add) const
      {
	timeval T = Time;

	T.tv_usec += Add.Time.tv_usec;
	T.tv_sec += Add.Time.tv_sec;
	return( AccurateTime( T ) );
      }
    AccurateTime operator+=(const AccurateTime &Add)
      {
	Time.tv_usec += Add.Time.tv_usec;
	Time.tv_sec += Add.Time.tv_sec;
	return( AdjustUnits() );
      }
    int operator<(const AccurateTime &Compare)
      {
	return( timercmp( &Time, &Compare.Time, < ) );
      }

    operator int() const
      {
	/*
	 * Cast the time to seconds, with rounding.
	 */

	int Seconds = int( Time.tv_sec );
	if (Time.tv_usec > 500000)
	    Seconds++;
	  else if (Time.tv_usec < -500000)
	    Seconds--;
	return( Seconds );
      }

    AccurateTime Mark() const
      {
	/*
	 * Return the difference between the current time and the time
	 * recorded by Start().
	 */

	struct timezone TZ;
	timeval T;
	gettimeofday( &T, &TZ );
	AccurateTime ReturnTime( T );
	ReturnTime -= *this;
	return( ReturnTime );
      }
    long Microseconds() const
      {
	return( Time.tv_usec );
      }
    long Seconds() const
      {
	return( Time.tv_sec );
      }
    void Start()
      {
	/*
	 * Record the current time.
	 */

	struct timezone TZ;
	gettimeofday( &Time, &TZ );
      }
  };


inline ostream &operator<<(ostream &S, AccurateTime AT)
  {
    return( S << AT.Seconds() << " sec " << AT.Microseconds() << " usec " );
  }


#endif
