#-----------------------------------------------------------------------------#
#		       Copyright (c) John F. Croix, 1997
#
#			      All Rights Reserved
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any non-commercial or internal-use-only purpose and without
# fee is hereby granted, provided that the above copyright notice appears in
# all copies and that both the copyright notice and this permission notice
# appear in supporting documentation, and all changes or alterations to the
# software and documentation are clearly documented.  Resale of any part of
# this software or documentation is prohibited without a license.  Licenses may
# be obtained by contacting the author, John F. Croix, at John.Croix@amd.com or
# journeyman@mail.utexas.edu.
#
# THE AUTHOR, JOHN F. CROIX, DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
# SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
# DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
# TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OF PERFORMANCE
# OF THIS SOFTWARE.
#-----------------------------------------------------------------------------#


import pipe_ext_type
import regex
import string
import types


#-----------------------------------------------------------------------------#
# This module contains the code necessary to implement Python object-oriented
# Expect-like processing.  The class Pipe is used to communicate with a
# process in background.  Patterns can be applied to the process to determine
# if a string match is found.  If so, the user can apply some action(s) based
# upon the matching pattern.
#.............................................................................#
# Available patterns:
#    o Pattern -- the template from which all "real" patterns are derived
#    o RegexPattern -- used to match compiled regular expressions
#
# We also have to have special types of patterns that do not really match
# via regular expression:
#    o EOF -- match if end-of-file found
#    o ERROR -- match if error discovered
#    o TIMEOUT -- match if no regular expression match found within a period
#          of time (user specified)
#    o FULL_BUFFER -- match if character buffer is full and must be shifted,
#          resulting in loss of characters
#    o DEFAULT -- match if nothing else matches
#
# Not yet implemented:
#    o NULL -- match on NULL character received
#-----------------------------------------------------------------------------#


BAD_OPEN_TYPE = "Invalid object type to open"


class Pattern:
    #
    # This class acts as a template for all other types of patterns which are
    # used to match the Pipe process.
    #

    def __init__(Self):
	#
	# Indicate that this pattern has not yet matched any character
	# string.
	#

	Self.Clear()

    def __repr__(Self):
	#
	# Default method when attempting to print the pattern.
	#

	return "Pattern --> <empty>"

    def Apply(Self, PipeProcess, BeforeMatch, MatchingText):
	#
	# BeforeMatch is the text that appears before the text matching the
	# pattern.  MatchingText is the text that matches the pattern.  This
	# function should be overridden to perform any actions on matching
	# the pattern.
	#
	# NOTE: Be sure to call this method from inheriting classes!
	#

	Self.MatchedPattern = 1
	Self.Leader = BeforeMatch
	Self.MatchingText = MatchingText

    def Clear(Self):
	#
	# Clear the flag indicating a match was found with this pattern
	#

	Self.MatchedPattern = 0
	Self.Leader = None
	Self.MatchingText = None

    def Match(Self, PipeProcess, Buffer):
	#
	# Override this method to test to see if the pattern matches the
	# contents of buffer.  Return the starting position of the match and
	# the length.  If there is no match, return -1 as the starting
	# position and 0 as the length.
	#

	return (-1, 0)

    def Matched(Self):
	#
	# Return the status of this instance.  Did this pattern match the
	# string?
	#

	return Self.MatchedPattern


class DEFAULT(Pattern):
    #
    # The default pattern always matches.
    #

    def __init__(Self):
	Pattern.__init__(Self)

    def __repr__(Self):
	return "Pattern --> DEFAULT"

    def Match(Self, PipeProcess, Buffer):
	#
	# We want to match at the beginning of the buffer for a length of 0.
	# This guarantees a match without eating characters from the buffer.
	# However, we do not want an asynchronous match to choose this one over
	# another, valid match.  Thus we return a length of 0.
	#

	return (0, 0)


class EOF(Pattern):
    #
    # The Pipe reached the end-of-file
    #

    def __init__(Self):
	Pattern.__init__(Self)

    def __repr__(Self):
	return "Pattern --> EOF"

    def Match(Self, PipeProcess, Buffer):
	#
	# Return a valid match if the PipeProcess hit the end of the input
	# stream.  Eat all characters in the buffer.
	#

	if PipeProcess.EOF:
	    return (0, len( Buffer ))
	return (-1, 0)


class ERROR(Pattern):
    #
    # This pattern is used to match in the event that the Pipe process
    # discovered an error.  The contents of the match is the error message
    # from the process.
    #

    def __init__(Self):
	Pattern.__init__(Self)

    def __repr__(Self):
	return "Pattern --> ERROR"

    def Apply(Self, PipeProcess, BeforeMatch, MatchingText):
	#
	# In this case, the matching text is the error message
	#

	Text = PipeProcess.ErrorMsg()
	Pattern.Apply( Self, PipeProcess, "", Text )

    def ErrorMsg(Self):
	#
	# Return the error message to the user
	#

	return (Self.MatchingText)

    def Match(Self, PipeProcess, Buffer):
	#
	# Match on error
	#

	if PipeProcess.Error:
	    return (0, 0)
	return (-1, 0)


class FULL_BUFFER(Pattern):
    #
    # This pattern is used to match the Pipe process if the buffer is full
    # and will be processed to make more room for characters.
    #

    def __init__(Self):
	Pattern.__init__(Self)

    def __repr__(Self):
	return "Pattern --> FULL_BUFFER"

    def Match(Self, PipeProcess, Buffer):
	#
	# Return a valid match if the PipeProcess buffer is full.  Eat all
	# characters in the buffer.
	#

	if PipeProcess.FullBuffer:
	    return (0, len( Buffer ))
	return (-1, 0)


class TIMEOUT(Pattern):
    #
    # The Pipe process had a timeout.  This matches that situation.
    #

    def __init__(Self):
	Pattern.__init__(Self)

    def __repr__(Self):
	return "Pattern --> TIMEOUT"

    def Match(Self, PipeProcess, Buffer):
	#
	# Return a valid match if the PipeProcess had a timeout.  Eat all
	# characters in the buffer.
	#

	if PipeProcess.Timeout:
	    return (0, 0)
	return (-1, 0)


class RegexPattern(Pattern):
    #
    # Template for a regular expression matching pattern.
    #

    def __init__(Self, Expression):
	Pattern.__init__(Self)
	Self.TextExpression = Expression
	Self.RegularExpression = regex.compile( Expression )

    def __repr__(Self):
	return "Pattern --> " + Self.TextExpression

    def group(Self, *Indices):
	#
	# Return the substring matched by the grouped subexpression
	#

	if Self.MatchedPattern:
	    return Self.RegularExpression.group( Indices )
	return None

    def regs(Self, Index):
	#
	# Return the stop/start index tuple for the substring matched by
	# the grouped subexpressions.
	#

	if Self.MatchedPattern:
	    return Self.RegularExpression.regs[ Index ]
	return (-1, -1)

    def Match(Self, PipeProcess, Buffer):
	#
	# See if the regular expression matches.  If so, return the starting
	# position and length.  Otherwise, return -1 and 0.
	#

	StartPos = Self.RegularExpression.search( Buffer )
	if (StartPos >= 0):
	    Length = Self.RegularExpression.match( Buffer, StartPos )
	else:
	    Length = 0
	return (StartPos, Length)


class RegexNocasePattern(RegexPattern):
    #
    # Template for a regular expression matching pattern in which there should
    # be no case distinction.  The easiest way to do this is to intercept
    # invocations and make all text lower case.
    #

    def __init__(Self, Expression):
	RegexPattern.__init__( Self, string.lower( Expression ) );
	Self.ActualPattern = Expression

    def __repr__(Self):
	return "Pattern --> " + Self.ActualPattern

    def Match(Self, PipeProcess, Buffer):
	#
	# Match against a lowercase version of the buffer
	#

	return RegexPattern.Match( Self, PipeProcess, \
				   string.lower( Buffer ) )

    
#-----------------------------------------------------------------------------#
# Here is the Pipe class.  It uses the C methods defined in pipe_ext_type
# (Pipe extension type) to actually do any of the processing.
#-----------------------------------------------------------------------------#


class Pipe:
    def __init__(Self):
	Self.AfterPatterns = ()
	Self.AfterLinePatterns = ()
	Self.BeforePatterns = ()
	Self.BeforeLinePatterns = ()
	Self.EchoToUser = 0
	Self.EOF = 0
	Self.Error = 0
	Self.PipeProcess = pipe_ext_type.pipe_ext_type()
	Self.FullBuffer = 0
	Self.Timeout = 0

    def BufferSize(Self):
	#
	# Return the size of the buffer that is used for regular expressions.
	#

	return Self.PipeProcess.BufferSize()

    def CheckAsync(Self, Buffer, Patterns):
	#
	# Match the contents of Buffer against the stored patterns.  Unlike
	# the CheckSync() function, all patterns are tested.  The pattern that
	# has the earliest start position is the one that is used.  If
	# more than one has the same start position, the longest of those
	# is the matching pattern.
	#
	# Return 1 if a match is made, 0 otherwise.
	#

	Earliest = None
	for Pat in Patterns:
	    (StartPosition, Length) = Pat.Match( Self, Buffer )
	    if StartPosition >= 0:

		#
		# If there is a match, check to see if any prior match is a
		# better "fit".  In other words, did this match occur later
		# in the string?  Is it a shorter match?
		#
		# The one special case to worry about is the DEFAULT case.
		# It will always match the earliest.  However, it should have
		# a length of 0 meaning it did not really match anything.
		#

		ReplaceEarliest = 0
		if Earliest != None:
		    if Length == 0:
			continue
		    if Earliest[ 1 ] == 0:
			ReplaceEarliest = 1
		    elif StartPosition < Earliest[ 0 ]:
			ReplaceEarliest = 1
		    elif StartPosition > Earliest[ 0 ]:
			continue
		    elif Length > Earliest[ 1 ]:
			ReplaceEarliest = 1
		else:
		    ReplaceEarliest = 1
		if ReplaceEarliest:
		    Earliest = (StartPosition, Length, Pat)
	return Earliest

    def CheckPatterns(Self, Method, Patterns, LineByLine=0):
	#
	# This code provides the skeleton for checking a buffer against one
	# or more regular expressions.  The matching method (the routine that
	# actually performs the regular expression matching) is named in the
	# Method parameter.
	#
	# If LineByLine is TRUE, we perform pattern matches against a single
	# line at a time.  Otherwise, match the patterns against the entire
	# process buffer.
	#

	if LineByLine:
	    Patterns = Self.BeforeLinePatterns + \
		       Patterns + Self.AfterLinePatterns
	else:
	    Patterns = Self.BeforePatterns + Patterns + Self.AfterPatterns
	Self.ResetPatterns( Patterns )
	Self.PipeProcess.ResetTimeoutTimer()
	FunctionPtr = getattr( Self, Method )
	while 1:
	    if LineByLine:
		Buffer = Self.PipeProcess.Line( Self.EchoToUser )
	    else:
		Buffer = Self.PipeProcess.Data()
	    for Condition in ["Normal", "FullBuffer", "Timeout", \
			      "EOF", "Error"]:
		#
		# We need to try different conditions.  In "Normal" mode,
		# we just check regular patterns (ie not FULL_BUFFER or
		# TIMEOUT).  In the other two, we set condition flags and
		# retry the same patterns to see if one clicks.
		#

		Self.EOF = 0
		Self.Error = 0
		Self.FullBuffer = 0
		Self.Timeout = 0
		if Condition == "FullBuffer":
		    if Self.PipeProcess.FullBuffer():
			Self.FullBuffer = 1
		    else:
			continue
		elif Condition == "Timeout":
		    if Self.PipeProcess.ProcessTimedOut():
			Self.Timeout = 1
		    else:
			continue
		elif Condition == "EOF":
		    if Self.PipeProcess.ProcessCompleted():
			Self.EOF = 1
		    else:
			continue
		elif Condition == "Error":
		    if Self.PipeProcess.Error():
			Self.Error = 1
		    else:
			continue

		#
		# Try to match each pattern using the specified Method (class
		# member function).  If one is found that matches, prune the
		# data from the buffer and do the pattern callback.
		#

		Found = FunctionPtr( Buffer, Patterns )
		if Found != None:
		    #
		    # Prune the data.  If we are going LineByLine, the
		    # data has already been spliced from the buffer.
		    #

		    (StartPosition, Length, Pat) = Found
		    BeforeMatch = Buffer[:StartPosition]
		    EndPosition = StartPosition + Length
		    Matching = Buffer[StartPosition:EndPosition]
		    if EndPosition and (not LineByLine):
			Self.PipeProcess.Splice( 0, EndPosition - 1 )

		    #
		    # Invoke the matching patterns callback.  Reset any
		    # non-error exception condition as we must have matched
		    # the pattern here.
		    #

		    Pat.Apply( Self, BeforeMatch, Matching )
		    if Condition != "Error":
			Self.PipeProcess.ResetTimeoutTimer()
			Self.PipeProcess.ResetError()
		    return

	    #
	    # If no match was found, get more characters from the buffer.
	    #
	    # There are special circumstances we need to worry about:
	    #   o If a timeout occurred, and there was no TIMEOUT pattern,
	    #     we need to reset the timeout timer so that the process
	    #     will continue to extract data.
	    #   o If reading LineByLine, the Read() is automatically performed
	    #     for us.  Thus, do *NOT* do it again here.
	    #

	    if Self.PipeProcess.ProcessTimedOut():
		Self.PipeProcess.ResetTimeoutTimer()
		Self.PipeProcess.ResetError()

	    if not LineByLine:
		Self.PipeProcess.Read( Self.EchoToUser )

    def CheckSync(Self, Buffer, Patterns):
	#
	# Match the contents of Buffer against the stored patterns.  The
	# patterns are tested in the order they were passed.  The first
	# one to match stops the matching process.
	#
	# Return 1 if a match is made, 0 otherwise.
	#

	for Pat in Patterns:
	    (StartPosition, Length) = Pat.Match( Self, Buffer )
	    if StartPosition >= 0:
		return (StartPosition, Length, Pat)
	return None

    def ClearBuffer(Self):
	#
	# Zero out the contents of the data buffer from the child process.
	#

	Self.PipeProcess.ClearBuffer()

    def Close(Self):
	#
	# Close communications with the child process
	#

	Self.PipeProcess.Close()

    def CloseLog(Self):
	#
	# Close any open log file
	#

	Self.PipeProcess.CloseLog()

    def ErrorMsg(Self):
	#
	# Return any error message from the process
	#

	return Self.PipeProcess.ErrorMsg()

    def LogUser(Self, Value):
	#
	# The user wants to specify whether output should be displayed to
	# the user or not.
	#

	Self.EchoToUser = Value

    def Match(Self, *Patterns):
	#
	# Call CheckPatterns with the CheckSync method.
	#

	Self.CheckPatterns( "CheckSync", Patterns )

    def MatchAfter(Self, *Patterns):
	#
	# Save the patterns that we should always match against after any
	# other passed patterns.
	#

	Self.AfterPatterns = Patterns

    def MatchAsync(Self, *Patterns):
	#
	# Call CheckPatterns with the CheckAsync method.
	#

	Self.CheckPatterns( "CheckAsync", Patterns )

    def MatchBefore(Self, *Patterns):
	#
	# Save the patterns that we should always match against before any
	# other passed patterns.
	#

	Self.BeforePatterns = Patterns

    def MatchLine(Self, *Patterns):
	#
	# Call CheckPatterns with the CheckSync method.
	#

	Self.CheckPatterns( "CheckSync", Patterns, 1 )

    def MatchLineAfter(Self, *Patterns):
	#
	# Save the patterns that we should always match against after any
	# other passed patterns in a line-by-line match.
	#

	Self.AfterLinePatterns = Patterns

    def MatchLineAsync(Self, *Patterns):
	#
	# Call CheckPatterns with the CheckAsync method.
	#

	Self.CheckPatterns( "CheckAsync", Patterns, 1 )

    def MatchLineBefore(Self, *Patterns):
	#
	# Save the patterns that we should always match against before any
	# other passed patterns.
	#

	Self.BeforeLinePatterns = Patterns

    def Open(Self, Argument):
	#
	# Argument should be a string, in which case it represents a program
	# invocation, or it should be a file.  Anything else is an error.
	#
	# If it is a string, open the passed process.  Argument should be a
	# single string that contains the complete command, including arguments
	# (ie "ls -al *.py")
	#
	# If it is a file, we need to open that files file descriptor.
	#

	if type( Argument ) == types.StringType:
	    Self.PipeProcess.Open( Argument )
	elif type( Argument ) == types.FileType:
	    Self.PipeProcess.OpenFD( Argument.fileno() )
	else:
	    raise BAD_OPEN_TYPE

    def OpenLog(Self, LogFileName, LogFileMode):
	#
	# Open a log file with the specified name and specified mode.
	#

	return Self.PipeProcess.OpenLog( LogFileName, LogFileMode )

    def ResetPatterns(Self, Patterns):
	#
	# Reset the match indicator in each pattern
	#

	for Pat in Patterns:
	    Pat.Clear()

    def SetBufferSize(Self, NewSize):
	#
	# Set the size of the buffer that is used for regular expressions.
	#

	Self.PipeProcess.SetBufferSize( NewSize )

    def SetTimeout(Self, TimeoutPeriod):
	#
	# Set the timeout period to the passed value
	#

	Self.PipeProcess.SetTimeout( TimeoutPeriod )

    def STTY(Self, Command):
	#
	# Send the process' term a stty command
	#

	Self.PipeProcess.STTY( Command )

    def Timeout(Self):
	#
	# Return the current timeout period for the process
	#

	return Self.PipeProcess.Timeout()

    def Wait(Self):
	#
	# Wait for the child process to finish.  Return the status of the
	# C wait() function.
	#

	return Self.PipeProcess.Wait()

    def Write(Self, Command):
	#
	# Send Command to the process
	#

	Self.PipeProcess.Write( Command )

