/*****************************************************************************
		       Copyright (c) John F. Croix, 1997

			      All Rights Reserved

  Permission to use, copy, modify, and distribute this software and its
  documentation for any non-commercial or internal-use-only purpose and without
  fee is hereby granted, provided that the above copyright notice appears in
  all copies and that both the copyright notice and this permission notice
  appear in supporting documentation, and all changes or alterations to the
  software and documentation are clearly documented.  Resale of any part of
  this software or documentation is prohibited without a license.  Licenses may
  be obtained by contacting the author, John F. Croix, at John.Croix@amd.com or
  journeyman@mail.utexas.edu.

  THE AUTHOR, JOHN F. CROIX, DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
  SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR
  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
  DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
  TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OF PERFORMANCE
  OF THIS SOFTWARE.
 *****************************************************************************/


extern "C" 
  {
#include <Python.h>
  }

#include "pipe_process.h"
#include <stdio.h>
#include <string.h>


struct PipeObject
  {
    PyObject_HEAD
    PipeProcess *Object;
  };


/*****************************************************************************
 * Declare routines in this file
 *****************************************************************************/


static PyObject *PipeBufferSize(PipeObject *EO, PyObject *Args);
static PyObject *PipeClearBuffer(PipeObject *EO, PyObject *Args);
static PyObject *PipeCloseLog(PipeObject *EO, PyObject *Args);
static PyObject *PipeData(PipeObject *EO, PyObject *Args);
static PyObject *PipeError(PipeObject *EO, PyObject *Args);
static PyObject *PipeErrorMsg(PipeObject *EO, PyObject *Args);
static PyObject *PipeFullBuffer(PipeObject *EO, PyObject *Args);
static PyObject *PipeLength(PipeObject *EO, PyObject *Args);
static PyObject *PipeLine(PipeObject *EO, PyObject *Args);
static PyObject *PipeOpenLog(PipeObject *EO, PyObject *Args);
static PyObject *PipeRemoveNulls(PipeObject *EO, PyObject *Args);
static PyObject *PipeSetBufferSize(PipeObject *EO, PyObject *Args);
static PyObject *PipeSplice(PipeObject *EO, PyObject *Args);

static PyObject *PipeProcessCompleted(PipeObject *EO, PyObject *Args);
static PyObject *PipeProcessTimedOut(PipeObject *EO, PyObject *Args);
static PyObject *PipeResetError(PipeObject *EO, PyObject *Args);
static PyObject *PipeResetTimeoutTimer(PipeObject *EO, PyObject *Args);

static PyObject *PipeClose(PipeObject *EO, PyObject *Args);
static PyObject *PipeOpen(PipeObject *EO, PyObject *Args);
static PyObject *PipeOpenFD(PipeObject *EO, PyObject *Args);
static PyObject *PipeRead(PipeObject *EO, PyObject *Args);
static PyObject *PipeSetTimeout(PipeObject *EO, PyObject *Args);
static PyObject *PipeSTTY(PipeObject *EO, PyObject *Args);
static PyObject *PipeTimeout(PipeObject *EO, PyObject *Args);
static PyObject *PipeWait(PipeObject *EO, PyObject *Args);
static PyObject *PipeWrite(PipeObject *EO, PyObject *Args);

static PipeObject *NewPipeObject();
static void DeletePipeObject(PipeObject *EO);
static int PrintPipeObject(PipeObject *EO);
static PyObject *GetAttrPipeObject(PipeObject *EO, const char *Name);

static PyObject *AllocatePipeObject(PyObject *EO, PyObject *Args);


/*****************************************************************************
 * Type descriptors.  See Python Include/Object.h for list of settings.
 *****************************************************************************/


static PyTypeObject PipeType =
{ PyObject_HEAD_INIT(&PyType_Type)
  0,                                  /* ob_size      */
  "PipeExtensionType",                /* tp_name      */
  sizeof(PipeObject),                 /* tp_basicsize */
  0,                                  /* tp_itemsize  */

  (destructor) DeletePipeObject,      /* tp_dealloc @ ref-count = 0 */
  (printfunc) PrintPipeObject,        /* tp_print   */
  (getattrfunc) GetAttrPipeObject,    /* tp_getattr */
  (setattrfunc) 0,                    /* tp_setattr */
  (cmpfunc) 0,                        /* tp_compare */
  (reprfunc) 0,                       /* tp_repr    */

  0,                                  /* tp_as_number   */
  0,                                  /* tp_as_seauence */
  0,                                  /* tp_as_mapping  */

  (hashfunc) 0,                       /* tp_hash */
  (ternaryfunc) 0,                    /* tp_call */
  (reprfunc) 0,                       /* tp_str  */
  0, 0, 0, 0, 0                       /* Space for future expansion */
};


#define IS_PIPE_OBJECT(O)  ((O)->ob_type == &PipeType)


/*****************************************************************************
 * Buffer manipulation
 *****************************************************************************/


PyObject *PipeBufferSize(PipeObject *EO, PyObject *Args)
  {
    /*
     * Return the size (capacity) of the buffer.
     */

    if (!PyArg_ParseTuple( Args, "" ))
        return NULL;
    return( Py_BuildValue( "i", EO->Object->Size() ) );
  }


PyObject *PipeClearBuffer(PipeObject *EO, PyObject *Args)
  {
    /*
     * Clear the buffer
     */

    if (!PyArg_ParseTuple( Args, "" ))
        return NULL;
    EO->Object->Clear();
    return( Py_BuildValue( "i", 1 ) );
  }


PyObject *PipeCloseLog(PipeObject *EO, PyObject *Args)
  {
    /*
     * Close any open log file
     */

    if (!PyArg_ParseTuple( Args, "" ))
        return NULL;
    EO->Object->CloseLog();
    return( Py_BuildValue( "i", 1 ) );
  }


PyObject *PipeData(PipeObject *EO, PyObject *Args)
  {
    /*
     * Return the string in the buffer
     */

    if (!PyArg_ParseTuple( Args, "" ))
        return NULL;
    return( Py_BuildValue( "s", EO->Object->Data() ) );
  }


PyObject *PipeError(PipeObject *EO, PyObject *Args)
  {
    /*
     * Return the string that represents any error messages
     */

    if (!PyArg_ParseTuple( Args, "" ))
        return NULL;
    int HasError = 0;
    if (EO->Object->ErrorValue() != PipeProcess::OK)
        HasError = 1;
    return( Py_BuildValue( "i", HasError ) );
  }


PyObject *PipeErrorMsg(PipeObject *EO, PyObject *Args)
  {
    /*
     * Return the string that represents any error messages
     */

    if (!PyArg_ParseTuple( Args, "" ))
        return NULL;
    return( Py_BuildValue( "s", EO->Object->ErrorMsg() ) );
  }


PyObject *PipeFullBuffer(PipeObject *EO, PyObject *Args)
  {
    /*
     * Is the buffer full (ie will characters be lost from the buffer at
     * the next read)?
     */

    if (!PyArg_ParseTuple( Args, "" ))
        return NULL;
    return( Py_BuildValue( "i", EO->Object->FullBuffer() ) );
  }


PyObject *PipeLength(PipeObject *EO, PyObject *Args)
  {
    /*
     * Return the length of the string in the buffer.
     */

    if (!PyArg_ParseTuple( Args, "" ))
        return NULL;
    return( Py_BuildValue( "i", EO->Object->Length() ) );
  }


PyObject *PipeLine(PipeObject *EO, PyObject *Args)
  {
    /*
     * Return the string in the buffer
     */

    int Echo;
    if (!PyArg_ParseTuple( Args, "i", &Echo ))
        return NULL;
    const char *Buffer = EO->Object->Line( Echo );
    return( Py_BuildValue( "s", Buffer ? Buffer : "" ) );
  }


PyObject *PipeOpenLog(PipeObject *EO, PyObject *Args)
  {
    /*
     * Open a (new) log file.
     */

    char *LogFile, *Mode;
    if (!PyArg_ParseTuple( Args, "ss", &LogFile, &Mode ))
        return NULL;
    return( Py_BuildValue( "i", EO->Object->OpenLog( LogFile, Mode ) ) );
  }



PyObject *PipeProcessCompleted(PipeObject *EO, PyObject *Args)
  {
    /*
     * Determine if the process finished.  If so, return 1.  Otherwise,
     * return 0.
     */

    if (!PyArg_ParseTuple( Args, "" ))
        return NULL;
    int FoundEOF = EO->Object->ErrorValue()==PipeProcess::EOF_FOUND ? 1 : 0;
    return( Py_BuildValue( "i", FoundEOF ) );
  }


PyObject *PipeProcessTimedOut(PipeObject *EO, PyObject *Args)
  {
    /*
     * Determine if the process timed out.  If so, return 1.  Otherwise,
     * return 0.
     */

    if (!PyArg_ParseTuple( Args, "" ))
        return NULL;
    int HadTimeOut = EO->Object->ErrorValue()==PipeProcess::TIMEOUT ? 1 : 0;
    return( Py_BuildValue( "i", HadTimeOut ) );
  }


PyObject *PipeRemoveNulls(PipeObject *EO, PyObject *Args)
  {
    /*
     * Remove any NULL characters from the string.
     */

    if (!PyArg_ParseTuple( Args, "" ))
        return NULL;
    EO->Object->RemoveNulls();
    return( Py_BuildValue( "i", 1 ) );
  }


PyObject *PipeResetError(PipeObject *EO, PyObject *Args)
  {
    /*
     * Reset any error condition flag.
     */

    if (!PyArg_ParseTuple( Args, "" ))
        return NULL;
    EO->Object->ResetError();
    return( Py_BuildValue( "i", 1 ) );
  }


PyObject *PipeResetTimeoutTimer(PipeObject *EO, PyObject *Args)
  {
    /*
     * Reset the timer used to determine timeout conditions.
     */

    if (!PyArg_ParseTuple( Args, "" ))
        return NULL;
    EO->Object->ResetTimeoutTimer();
    return( Py_BuildValue( "i", 1 ) );
  }


PyObject *PipeSetBufferSize(PipeObject *EO, PyObject *Args)
  {
    /*
     * Set the capacity of the buffer
     */

    unsigned Size;
    if (!PyArg_ParseTuple( Args, "i", &Size ))
        return NULL;
    return( Py_BuildValue( "i", EO->Object->Size( Size ) ) );
  }


PyObject *PipeSplice(PipeObject *EO, PyObject *Args)
  {
    /*
     * Splice contents out of the array.  Return the contents spliced.
     */

    unsigned I1, I2;
    if (!PyArg_ParseTuple( Args, "ii", &I1, &I2 ))
        return NULL;

    PyObject *ReturnObject = NULL;
    char *Buffer = new char[ I2 - I1 + 2 ];
    if (EO->Object->Splice( I1, I2, Buffer ))
	{
	  ReturnObject = Py_BuildValue( "s", Buffer );
	  delete Buffer;
	}
    return( ReturnObject );
  }


/*****************************************************************************
 * Process communication
 *****************************************************************************/


PyObject *PipeClose(PipeObject *EO, PyObject *Args)
  {
    /*
     * Close the process
     */

    if (!PyArg_ParseTuple( Args, "" ))
        return NULL;
    EO->Object->Close();
    return( Py_BuildValue( "i", 1 ) );
  }


PyObject *PipeOpen(PipeObject *EO, PyObject *PArgs)
  {
    /*
     * Start a process.
     */

    char *Invocation;
    if (!PyArg_ParseTuple( PArgs, "s", &Invocation ))
        return NULL;

    /*
     * We need to break up Invocation so that we can pass it as a program
     * name followed by the program arguments.
     */

    char *Args = strdup( Invocation );
    int NumArgs = 1;
    strtok( Args, " \t" );
    while (strtok( 0, " \t" ))
      NumArgs++;

    strcpy( Args, Invocation );
    const char **Arguments = new const char *[NumArgs + 1];
    Arguments[ 0 ] = strtok( Args, " \t" );
    NumArgs = 1;
    while (Arguments[ NumArgs++ ] = strtok( 0, " \t" ));
    EO->Object->Open( Arguments[ 0 ], Arguments );

    /*
     * Free used memory.  Return status code.
     */

    delete[] Arguments;
    free( Args );
    return( Py_BuildValue( "i", 1 ) );
  }


PyObject *PipeOpenFD(PipeObject *EO, PyObject *PArgs)
  {
    /*
     * Set up a file descriptor to interact with.
     */

    int FileDescriptor;
    if (!PyArg_ParseTuple( PArgs, "i", &FileDescriptor ))
        return NULL;
    EO->Object->Open( FileDescriptor );
    return( Py_BuildValue( "i", 1 ) );
  }


PyObject *PipeRead(PipeObject *EO, PyObject *Args)
  {
    /*
     * Read from the child process.  Return the number of bytes read.
     */

    int EchoToUser;
    if (!PyArg_ParseTuple( Args, "i", &EchoToUser ))
        return NULL;
    return( Py_BuildValue( "i", EO->Object->Read( EchoToUser ) ) );
  }


PyObject *PipeTimeout(PipeObject *EO, PyObject *Args)
  {
    /*
     * Return the timeout period.
     */

    if (!PyArg_ParseTuple( Args, "" ))
        return NULL;
    return( Py_BuildValue( "i", EO->Object->Timeout() ) );
  }


PyObject *PipeSetTimeout(PipeObject *EO, PyObject *Args)
  {
    /*
     * Set the timeout period
     */

    int NewTimeout;
    if (!PyArg_ParseTuple( Args, "i", &NewTimeout ))
        return NULL;
    return( Py_BuildValue( "i", EO->Object->Timeout( NewTimeout ) ) );
  }



PyObject *PipeSTTY(PipeObject *EO, PyObject *Args)
  {
    /*
     * Send a "stty" command to the process
     */

    char *STTYCommand;
    if (!PyArg_ParseTuple( Args, "s", &STTYCommand ))
        return NULL;
    EO->Object->STTY( STTYCommand );
    return( Py_BuildValue( "i", 1 ) );
  }


PyObject *PipeWait(PipeObject *EO, PyObject *Args)
  {
    /*
     * Wait for the process to terminate
     */

    if (!PyArg_ParseTuple( Args, "" ))
        return NULL;
    return( Py_BuildValue( "i", EO->Object->Wait() ) );
  }


PyObject *PipeWrite(PipeObject *EO, PyObject *Args)
  {
    /*
     * Write out the passed string.
     */

    char *WriteData;
    if (!PyArg_ParseTuple( Args, "s", &WriteData ))
        return NULL;
    return( Py_BuildValue( "i", EO->Object->Write( WriteData ) ) );
  }


/*****************************************************************************
 * Define the methods available for Python to use
 *****************************************************************************/


static PyMethodDef PipeMethods[] =
{ { "BufferSize", PyCFunction( PipeBufferSize ), 1 },
  { "ClearBuffer", PyCFunction( PipeClearBuffer ), 1 },
  { "Close", PyCFunction( PipeClose ), 1 },
  { "CloseLog", PyCFunction( PipeCloseLog ), 1 },
  { "Data", PyCFunction( PipeData ), 1 },
  { "Error", PyCFunction( PipeError ), 1 },
  { "ErrorMsg", PyCFunction( PipeErrorMsg ), 1 },
  { "FullBuffer", PyCFunction( PipeFullBuffer ), 1 },
  { "Length", PyCFunction( PipeLength ), 1 },
  { "Line", PyCFunction( PipeLine ), 1 },
  { "Open", PyCFunction( PipeOpen ), 1 },
  { "OpenFD", PyCFunction( PipeOpenFD ), 1 },
  { "OpenLog", PyCFunction( PipeOpenLog ), 1 },
  { "ProcessCompleted", PyCFunction( PipeProcessCompleted ), 1 },
  { "ProcessTimedOut", PyCFunction( PipeProcessTimedOut ), 1 },
  { "Read", PyCFunction( PipeRead ), 1 },
  { "RemoveNulls", PyCFunction( PipeRemoveNulls ), 1 },
  { "ResetError", PyCFunction( PipeResetError ), 1 },
  { "ResetTimeoutTimer", PyCFunction( PipeResetTimeoutTimer ), 1 },
  { "SetBufferSize", PyCFunction( PipeSetBufferSize ), 1 },
  { "SetTimeout", PyCFunction( PipeSetTimeout ), 1 },
  { "Splice", PyCFunction( PipeSplice ), 1 },
  { "STTY", PyCFunction( PipeSTTY ), 1 },
  { "Timeout", PyCFunction( PipeTimeout ), 1 },
  { "Wait", PyCFunction( PipeWait ), 1 },
  { "Write", PyCFunction( PipeWrite ), 1 },
  { NULL, NULL } };


/*****************************************************************************
 * Basic operations
 *    o new
 *    o delete
 *    o print
 *    o any type of <instance>.<name> reference
 *****************************************************************************/


PipeObject *NewPipeObject()
  {
    /*
     * Allocate a new structure for Python to refer to an PipeProcess.
     * Then allocate the PipeProcess under it.
     */

    PipeObject *EO = PyObject_NEW( PipeObject, &PipeType );
    if (EO)
        EO->Object = new PipeProcess();
    return( EO );
  }


void DeletePipeObject(PipeObject *EO)
  {
    /*
     * Call the PipeProcess destructor.  Then free Python's memory
     * allocation for the data structure.  This is called when the reference
     * count for EO becomes 0.
     */

    delete EO->Object;
    PyMem_DEL( EO );
  }


int PrintPipeObject(PipeObject *EO)
  {
    /*
     * Print the PipeProcess data.
     */

    return( 0 );
  }


PyObject *GetAttrPipeObject(PipeObject *EO, const char *Name)
  {
    /*
     * Invoked when <instance>.name is referenced.
     *
     * Check for the following:
     *    o Variable references
     *    o List of variable names in the class
     *    o A class method
     *
     * Currently, there are no variables.  Go directly to the class methods.
     */

    return( Py_FindMethod( PipeMethods, (PyObject *) EO, (char *) Name ) );
  }


/*****************************************************************************
 * Module logic
 *****************************************************************************/


PyObject *AllocatePipeObject(PyObject *EO, PyObject *Args)
  {
    /*
     * Allocate a new PipeObject.  On "x = Pipe.Pipe()" EO is unused
     */

    if (!PyArg_ParseTuple( Args, "" ))
        return NULL;
    return( (PyObject *) NewPipeObject() );
  }


static PyMethodDef PipeExtensionType_Methods[] =
{ { "pipe_ext_type", AllocatePipeObject, 1 },
  { NULL, NULL } };


extern "C" void initpipe_ext_type();

static PyObject *ErrorObject;

void initpipe_ext_type()
  {
    PyObject *Module = Py_InitModule( "pipe_ext_type",
				      PipeExtensionType_Methods );
    PyObject *Dict = PyModule_GetDict( Module );
    ErrorObject = Py_BuildValue( "s", "pipe_ext_type.error" );
    
    PyDict_SetItemString( Dict, "error", ErrorObject );
    if (PyErr_Occurred())
        Py_FatalError( "can't initialize PipeExtensionType module" );
  }

