/*****************************************************************************
		       Copyright (c) John F. Croix, 1997

			      All Rights Reserved

  Permission to use, copy, modify, and distribute this software and its
  documentation for any non-commercial or internal-use-only purpose and without
  fee is hereby granted, provided that the above copyright notice appears in
  all copies and that both the copyright notice and this permission notice
  appear in supporting documentation, and all changes or alterations to the
  software and documentation are clearly documented.  Resale of any part of
  this software or documentation is prohibited without a license.  Licenses may
  be obtained by contacting the author, John F. Croix, at John.Croix@amd.com or
  journeyman@mail.utexas.edu.

  THE AUTHOR, JOHN F. CROIX, DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
  SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR
  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
  DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
  TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OF PERFORMANCE
  OF THIS SOFTWARE.
 *****************************************************************************/


#include "accurate_time.h"
#include "pipe_process.h"
#include <ctype.h>
#include <errno.h>
#include <memory.h>
#include <signal.h>
#include <string.h>
#ifdef HAVE_PTYTRAP
#include <sys/ptyio.h>
#endif
#include <unistd.h>


/*****************************************************************************
 *----------------------------------------------------------------------------
 *		     V A R I A B L E   D I C T I O N A R Y
 *----------------------------------------------------------------------------
 * o Buffer --> Holds data read from the input stream
 * o BufferCursor --> Indicates the position in the string buffer which is to
 *   get the next character from the input stream.
 * o BufferLimit --> The number of characters allocated for Buffer.
 * o BufferSize --> The size of the dynamic memory allocation for Buffer[].
 *   This is the logical size.  The actual size is in BufferLimit.
 * o ErrorCode --> In the event of an error, this variable holds the value
 *   that was placed into errno by the system.
 * o LineBuffer --> Contains contents of one line of data (terminating in
 *   "\n") from the process.
 * o PTY --> Instance to hold process data.
 * o Status --> Holds any internal error condition
 * o ReadTimeout --> The number of seconds to wait for data from the process.
 *****************************************************************************/


PipeProcess::PipeProcess()
  {
    /*
     * Constructor.  Initialize class local variables.
     */

    Init();
  }


PipeProcess::PipeProcess(const char *Process, const char *Arguments[])
  {
    /*
     * Constructor.  Initialize class local variables.
     */

    Init();
    Open( Process, Arguments );
  }


PipeProcess::PipeProcess(int FileDescriptor)
  {
    /*
     * Constructor.  Initialize class local variables.
     */

    Init();
    Open( FileDescriptor );
  }


PipeProcess::~PipeProcess()
  {
    /*
     * Return memory to the OS.
     */

    Close();
    CloseLog();
    delete Buffer;
    delete LineBuffer;
  }


void PipeProcess::AlarmHandler(int)
  {
    /*
     * This alarm handler is used for read timeouts.  If the alarm goes
     * off, this routine is invoked.  We then jump past the read using
     * a C longjmp().
     */

    signal( SIGALRM, AlarmHandler );
    longjmp( JumpBuffer(), 1 );
  }


void PipeProcess::Clear()
  {
    /*
     * Clear the buffer.
     */

    BufferCursor = 0;
    Buffer[ BufferCursor ] = '\0';
    delete LineBuffer;
    LineBuffer = NULL;
  }


void PipeProcess::Close()
  {
    /*
     * Close the PTY
     */

    PTY.Close();
    Status = PTY_ERROR;
  }


void PipeProcess::CloseLog()
  {
    /*
     * Close the log file, if one is open.
     */

    if (LogFile)
        fclose( LogFile );
  }


const char *PipeProcess::Data() const
  {
    /*
     * Return the data in the process buffer.
     */

    return( Buffer );
  }


PipeProcess::ErrorStatus PipeProcess::ErrorValue() const
  {
    /*
     * Return the internal error code.
     */

    return( Status );
  }


const char *PipeProcess::ErrorMsg() const
  {
    /*
     * Return an error message based on the error code.
     */

    const char *Msg;
    switch (Status)
      {
	case EOF_FOUND:
	  Msg = "End-of-file found when reading from child process.";
	  break;
	case ERROR_FOUND:
	  Msg = strerror( ErrorCode );
	  break;
	case LOG_ERROR:
	  Msg = "Error opening log file.";
	  break;
	case OK:
	  Msg = "No errors detected.";
	  break;
	case PTY_ERROR:
	  Msg = PTY.ErrorMsg();
	  break;
	case RANGE_ERROR:
	  Msg = "Attempt to access past end of buffer.";
	  break;
	case TIMEOUT:
	  Msg = "Timeout occurred while waiting for input from child process.";
	  break;
	default:
	  Msg = "Unknown error.";
	  break;
      }
    return( Msg );
  }


int PipeProcess::FullBuffer() const
  {
    /*
     * Return 1 if the buffer has maxed out (the next read will cause a
     * truncation of buffer contents).
     */

    return( (BufferCursor + 1) == BufferSize ? 1 : 0 );
  }


void PipeProcess::Init()
  {
    /*
     * Initialize class local variables.
     */
    
    BufferCursor = 0;
    BufferLimit = 2 * DEFAULT_BUFFER_SIZE;
    BufferSize = 2 * DEFAULT_BUFFER_SIZE;
    Buffer = new char[ BufferLimit ];
    ErrorCode = 0;
    LineBuffer = NULL;
    LogFile = 0;
    Status = OK;
    ReadTimeout = DEFAULT_TIMEOUT;

    /*
     * Initially, the buffer contains no data.
     */

    Buffer[ BufferCursor ] = '\0';
    Timer = ReadTimeout;
  }


jmp_buf &PipeProcess::JumpBuffer()
  {
    /*
     * This static member function is used because some compilers don't
     * support static class variables (notably GCC).  This method does
     * work, even though it's clumsy.
     */

    static jmp_buf Buffer;

    return( Buffer );
  }


unsigned PipeProcess::Length() const
  {
    /*
     * Return the number of characters in the buffer.
     */

    return( BufferCursor );
  }


const char *PipeProcess::Line(int Echo)
  {
    /*
     * Get one line of data from the Buffer (terminating in a "\n").  If the
     * Buffer does not have one, read more data into the buffer and try again.
     * The second time, we either find a "\n" or hit the end of the buffer.
     */

    int Found = -1;
    while ((Status == OK) && (Found == -1))
      {
	for (unsigned I = 0; (Found == -1) && (I < BufferCursor); I++)
	  if (Buffer[ I ] == '\n')
	      Found = I;

	if (Found == -1)
	    Read( Echo );
      }
    if (Status != OK)
        return( NULL );

    /*
     * At this point we know that Found >= 0.  Copy that many characters into
     * LineBuffer.  Then pull that many characters off of the start of Buffer.
     */

    delete LineBuffer;
    LineBuffer = new char[ Found + 1 ];
    memcpy( LineBuffer, Buffer, Found );
    if (Found != -1)
        {
	  BufferCursor -= Found;
	  memcpy( Buffer, Buffer + Found + 1, BufferCursor );
	  BufferCursor--;
	}

    /*
     * LineBuffer may have some control characters at the end of it (line
     * '\r').  We need to remove these control characters and reset the end
     * of the string.
     */

    for (Found--; (Found >= 0) && iscntrl( LineBuffer[ Found ] ); Found--);
    LineBuffer[ Found + 1 ] = '\0';
    return( LineBuffer );
  }


void PipeProcess::Open(int FileDescriptor)
  {
    /*
     * Set up interaction with the passed file descriptor.
     */

    if (PTY.Open( FileDescriptor ) != PipePTY::OK)
        Status = PTY_ERROR;
  }


void PipeProcess::Open(const char *Process, const char *Arguments[])
  {
    /*
     * Create the named process.
     */

    if (PTY.Open( Process, Arguments ) != PipePTY::OK)
        Status = PTY_ERROR;
  }


int PipeProcess::OpenLog(const char *Name, const char *Mode)
  {
    /*
     * Open a log file for the process.
     */

    if (Status != OK)
        return( 0 );

    CloseLog();
    LogFile = fopen( Name, Mode );
    if (!LogFile)
        {
	  Status = LOG_ERROR;
	  return( 0 );
	}

    /*
     * Output anything already in the buffer.  This is important if the user
     * decides to open a log file at some intermediate point of communication
     * with the background process.  Consider the following:
     *        - open a process
     *        - perform some actions
     *        - on some pattern, open a log file
     *        - more processing
     * In this case, the buffer could contain more data after the match that
     * the user wanted captured.  There is no way to tell whether or not this
     * is the case before the fact.  Thus, to play it safe, save any data
     * currently in the buffer.
     */

    fwrite( Buffer, sizeof( char ), BufferCursor, LogFile );
    return( 1 );
  }


unsigned PipeProcess::Read(int Echo)
  {
    /*
     * Read characters from the PipeProcess.
     *
     * NOTE: We don't do anything if there is a recorded error.
     */

    if (Status != OK)
        return( 0 );

    /*
     * Figure out how many characters to read.  If necessary, shift some data
     * out of the buffer to make room for data to read.
     */

    unsigned CharsToRead;
    do
      {
	CharsToRead = (BufferSize - BufferCursor - 1);
	if (!CharsToRead)
	    {
	      /*
	       * Shift the buffer to make room for more data.
	       */

	      int RemoveAmount = BufferCursor / 2;
	      memmove( Buffer, Buffer + RemoveAmount,
		       BufferCursor - RemoveAmount + 1 );
	      BufferCursor = BufferCursor - RemoveAmount;
	    }
      } while (!CharsToRead);
    ssize_t BytesRead = ReadFromProcess( CharsToRead, Buffer + BufferCursor );

    /*
     * Move the buffer cursor to the new position, if there wasn't a timeout.
     */

    if ((!BytesRead || (BytesRead == -1)) && (Status == OK))
        Status = EOF_FOUND;
      else
	{
	  /*
	   * If Echo is set, write the new data to the console.
	   * If performing logging, write the data to the log file
	   */

	  if (Echo)
	      fwrite( Buffer + BufferCursor, sizeof( char ),
		      BytesRead, stdout );
	  if (LogFile)
	      fwrite( Buffer + BufferCursor, sizeof( char ),
		      BytesRead, LogFile );
	  BufferCursor += int( BytesRead );
	  Buffer[ BufferCursor ] = '\0';
	}
    return( unsigned( BytesRead ) );
  }


ssize_t PipeProcess::ReadFromProcess(unsigned CharsToRead, char *CharBuffer)
  {
    /*
     * Read characters from the PipeProcess.
     *
     * Assume a timeout condition.  If we actually read all of the bytes,
     * reset the error condition.
     */

    Status = TIMEOUT;
    ssize_t BytesRead = 0;
    if (ReadTimeout > 0)
	{
	  signal( SIGALRM, AlarmHandler );
	  alarm( int( Timer ) );
	}

    if (!setjmp( JumpBuffer() ))
	{
	  /*
	   * We've set the jump point.  Now read from the process stream.
	   */

	  AccurateTime AT;
	  AT.Start();
#ifndef HAVE_PTYTRAP
	  /*
	   * The problem with this section of code is that it may require
	   * a timeout condition before we can determine that an EOF occurred.
	   * On a timeout, a 0 value will be returned.  However, if timeouts
	   * have been disabled, we may never be able to determine EOF.
	   */

	  while (!(BytesRead = read( PTY.ID(), CharBuffer, CharsToRead )) &&
		 !PTY.IsProcess());
#else
	  /*
	   * This code is much better at handling EOF conditions.  By using
	   * the select() function, we can determine when the child process
	   * terminated almost immediately.  No timeout condition is required.
	   */

	  int Done = 0;
	  do
	    {
	      fd_set ReadFDs;

	      FD_ZERO( &ReadFDs );
	      FD_SET( PTY.ID(), &ReadFDs );
	      fd_set ExceptFDs = ReadFDs;

	      int ConditionCode = select( PTY.ID() + 1, (int *) &ReadFDs,
					  NULL, (int *) &ExceptFDs, NULL );
	      if (ConditionCode == -1)
		  {
		    /*
		     * According to the Expect code "window refreshes trigger
		     * EINTR, ignore".
		     */

		    if (errno == EINTR)
		        continue;
		  }
	      if (FD_ISSET( PTY.ID(), &ReadFDs))
		  {
		    /*
		     * If there is data to read, do so.
		     */

		    BytesRead = read( PTY.ID(), CharBuffer, CharsToRead );
		    Done = 1;
		  }
		else if (FD_ISSET( PTY.ID(), &ExceptFDs ))
		  {
		    /*
		     * Caught an exception.  It may be an EOF.
		     */

		    request_info Info;

		    ioctl( PTY.ID(), TIOCREQCHECK, &Info );
		    if (Info.request == TIOCCLOSE)
		        {
			  /*
			   * Flag this condition as an EOF.  We'll mark the
			   * status variable later.
			   */

			  BytesRead = 0;
			  Done = 1;
			}
		      else
			{
			  /*
			   * According to the Expect documentation, an open()
			   * was trapped here.  Try again to read.
			   */

			  ioctl( PTY.ID(), TIOCREQSET, &Info );
			}
		  }
	    } while (!Done);
#endif

	  Timer -= AT.Mark();
	  Status = OK;
	}
      else
	{
	  /*
	   * A timeout occurred.
	   */

	  Timer = 0;
	}

    /*
     * Reset the alarm so that we won't jump back to the wrong spot and
     * force a timeout by mistake.
     */

    if (ReadTimeout > 0)
        alarm( 0 );
    return( BytesRead );
  }


void PipeProcess::RemoveNulls()
  {
    /*
     * Remove any NULL characters from the buffer.
     */

    if ((Status != OK) && (Status != TIMEOUT) && (Status != EOF_FOUND))
        return;

    int Source=0, Target=0;
    while (Source < BufferCursor)
      {
	if (!Buffer[ Source ])
	    Source++;
	  else if (Source != Target)
	    Buffer[ Target++ ] = Buffer[ Source++ ];
	  else Source++, Target++;
      }
    Buffer[ BufferCursor = Target ] = '\0';
  }


void PipeProcess::ResetError()
  {
    /*
     * Reset the error status for this process.
     */

    Status = OK;
    PTY.ErrorReset();
    if (PTY.ErrorValue() != PipePTY::OK)
        Status = PTY_ERROR;
  }


void PipeProcess::ResetTimeoutTimer()
  {
    /*
     * Reset the remaining time left for a pattern match to the default
     * timeout period.
     */

    Timer = ReadTimeout;
  }


unsigned PipeProcess::Size() const
  {
    /*
     * Return the size of the buffer.  Actually, we're returning half of
     * the size of the buffer.  As in Expect, we'll keep additional buffer
     * space around as the pathological pattern match may require 2x the
     * matching string size.
     */

    return( BufferSize / 2 );
  }


unsigned PipeProcess::Size(unsigned NewSize)
  {
    /*
     * Resize the buffer.  In reality, we won't necessarily resize as it
     * can take a lot of time to allocate new memory and copy everything over.
     * Instead, if we're making the buffer smaller, we'll only change the
     * "logical" buffer size.  If we're making it bigger, but it's still
     * smaller than the acutal buffer size, we'll increase the "logical"
     * limit.  Only if we're making it larger than anything we've already
     * allocated will we actually allocate memory and copy.
     *
     * First, deal with the case where we're making the buffer bigger.
     */

    NewSize *= 2;
    if (NewSize > BufferLimit)
	{
	  char *NewBuffer = new char[ NewSize ];
	  memcpy( NewBuffer, Buffer, BufferCursor + 1 );
	  delete[] Buffer;
	  BufferLimit = NewSize;
	  Buffer = NewBuffer;
	  return( NewSize / 2 );
	}

    /*
     * We're making the buffer smaller.  To do this, first decrease the
     * logical size of the buffer.  Then, if the buffer's cursor is past
     * this logical limit, we need to shift data so that the cursor is at
     * the end of the buffer.
     */

    BufferSize = NewSize;
    if (BufferCursor >= BufferSize)
	{
	  unsigned Offset = BufferCursor - BufferSize + 1;
	  memmove( Buffer, Buffer + Offset, BufferSize );
	}
    return( NewSize / 2 );
  }



unsigned PipeProcess::Splice(unsigned From, unsigned To, char *SpliceBuffer)
  {
    /*
     * Copy the contents between the characters located at From and
     * To into the passed buffer.  Remove this data from the internal
     * buffer.
     */

    if ((Status != OK) && (Status != TIMEOUT) && (Status != EOF_FOUND))
        return( 0 );
    if ((From > To) || (To > BufferCursor))
	{
	  Status = RANGE_ERROR;
	  return( 0 );
	}

    if (SpliceBuffer)
        {
	  memcpy( SpliceBuffer, Buffer + From, size_t( To - From + 1 ) );
	  SpliceBuffer[ To - From + 1 ] = '\0';
	}
    memmove( Buffer + From, Buffer + To + 1, size_t( BufferCursor - To ) );
    BufferCursor -= (To - From + 1);
    return( To - From + 1 );
  }


void PipeProcess::STTY(const char *Command)
  {
    /*
     * Pass a "stty" command to the PTY if it is not yet open.
     */

    if (Status == OK)
        PTY.STTY( Command );
  }


int PipeProcess::Timeout() const
  {
    /*
     * Return the timeout value (in seconds) that read must respond in.  Set
     * to 0 or less for unlimited wait.
     */

    return( ReadTimeout );
  }


int PipeProcess::Timeout(int NewTimeout)
  {
    /*
     * Set the timeout value (in seconds) that read must respond in.  Set
     * to 0 or less for unlimited wait.
     */

    return( ReadTimeout = NewTimeout);
  }


int PipeProcess::Wait()
  {
    /*
     * Wait for the process to finish.
     */

    if (Status == OK)
        return( PTY.Wait() );
    return( -1 );
  }


unsigned PipeProcess::Write(const char *Message, unsigned Len)
  {
    /*
     * Write Message to the process.  If Len != 0, write only Len bytes.
     * Otherwise, use strlen() to determine the number of bytes to write.
     */

    if ((Status != OK) && (Status != TIMEOUT))
        return( 0 );
    if (PTY.ErrorValue() != PipePTY::OK)
	{
	  Status = PTY_ERROR;
	  return( 0 );
	}

    if (!Len)
        Len = strlen( Message );
    return( write( PTY.ID(), Message, Len ) );
  }



