/*****************************************************************************
		       Copyright (c) John F. Croix, 1997

			      All Rights Reserved

  Permission to use, copy, modify, and distribute this software and its
  documentation for any non-commercial or internal-use-only purpose and without
  fee is hereby granted, provided that the above copyright notice appears in
  all copies and that both the copyright notice and this permission notice
  appear in supporting documentation, and all changes or alterations to the
  software and documentation are clearly documented.  Resale of any part of
  this software or documentation is prohibited without a license.  Licenses may
  be obtained by contacting the author, John F. Croix, at John.Croix@amd.com or
  journeyman@mail.utexas.edu.

  THE AUTHOR, JOHN F. CROIX, DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
  SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR
  CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
  DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
  TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OF PERFORMANCE
  OF THIS SOFTWARE.
 *****************************************************************************/


#include "pipe_pty.h"
#include <errno.h>
#include <expect.h>
#include <string.h>
#include <unistd.h>


PipePTY::PipePTY()
  {
    /*
     * Constructor.  Initialize class local variables.
     */

    ChildID = 0;
    SpawnID = -1;
    Status = PTY_UNINITIALIZED;
    SystemErrno = 0;
    TermState = 0;

    const char *SANE = "sane";
    STTYCommand = strcpy( new char[ strlen( SANE ) + 1 ], SANE );
  }


PipePTY::~PipePTY()
  {
    /*
     * Destructor.  Close open PTY or reset file descriptor.
     */

    Close();
    delete STTYCommand;
  }


void PipePTY::Close()
  {
    /*
     * If we have a value in TermState, we need to restore the state of
     * the file descriptor.  If not, we have a subprocess which needs to be
     * closed.
     */

    if (SpawnID != -1)
        {
	  if (TermState)
	      {
		tcsetattr( 0, TCSANOW, TermState );
		delete TermState;
		TermState = 0;
	      }
	    else close( SpawnID );
	  SpawnID = -1;
	  Status = PTY_CLOSED;
	}
  }


PipePTY::ErrorStatus PipePTY::ErrorValue() const
  {
    /*
     * Return the internal error code.
     */

    return( Status );
  }


const char *PipePTY::ErrorMsg() const
  {
    /*
     * Return an error message based on the error code.
     */

    const char *Msg;
    switch (Status)
      {
	case BAD_SPAWN:
	  Msg = "Could not spawn process.";
	  break;
	case BAD_STTY:
	  Msg = "Could not issue STTY after process opened.";
	  break;
	case OK:
	  Msg = "No errors detected.";
	  break;
	case PTY_CANNOT_WAIT:
	  Msg = "PTY must be closed before wait() can be issued.";
	  break;
	case PTY_CLOSED:
	  Msg = "PTY has been closed.";
	  break;
	case PTY_UNINITIALIZED:
	  Msg = "PTY has not been opened.";
	  break;
	case SYSTEM_ERROR:
	  Msg = strerror( SystemErrno );
	  break;
	case BAD_FILE_DES:
	  Msg = "Flow control on file descriptor cannot be set.";
	  break;
	default:
	  Msg = "Unknown error.";
	  break;
      }
    return( Msg );
  }


void PipePTY::ErrorReset()
  {
    /*
     * Reset the error status for this process.
     */

    if ((Status != PTY_UNINITIALIZED) && (Status != PTY_CLOSED))
        Status = OK;
  }


int PipePTY::ID() const
  {
    /*
     * Return the file descriptor to communicate with the process.
     */

    return( SpawnID );
  }


int PipePTY::IsProcess() const
  {
    /*
     * Return 1 if we are reading from a process, 0 otherwise.
     */

    return( TermState ? 0 : 1 );
  }


PipePTY::ErrorStatus PipePTY::Open(int FileDescriptor)
  {
    /*
     * Given the passed file descriptor (for example, stdin),
     * set up the file descriptor so that unbuffered interaction can
     * occur.
     */

    if ((SpawnID == -1) && (Status == PTY_UNINITIALIZED))
	{
	  int DoReset = 1;
	  TermState = new termios;

	  if (!tcgetattr( FileDescriptor, TermState ))
	      {
		termios NewTerm = *TermState;
//		NewTerm.c_lflag &= ~(ICANON|ECHO);
		NewTerm.c_lflag &= ~ICANON;
		DoReset = tcsetattr( FileDescriptor, TCSANOW, &NewTerm );
	      }

	  if (DoReset)
	      {
		Status = BAD_FILE_DES;
		delete TermState;
		TermState = 0;
	      }
	    else
	      {
		Status = OK;
		ChildID = getpid();
		SpawnID = FileDescriptor;
	      }
	}
    return( Status );
  }
    

PipePTY::ErrorStatus PipePTY::Open(const char *Program,
				   const char *Arguments[])
  {
    /*
     * Open a channel to the subprocess.
     */

    if ((SpawnID == -1) && (Status == PTY_UNINITIALIZED))
	{
	  extern char *exp_stty_init;
	  extern int exp_pid;

	  exp_stty_init = STTYCommand;
	  SpawnID = exp_spawnv( (char *) Program, (char **) Arguments );
	  if (SpawnID < 0)
	      Status = BAD_SPAWN;
	    else
	      {
		ChildID = pid_t( exp_pid );
		Status = OK;
	      }
	  exp_stty_init = 0;
	}
    return( Status );
  }


void PipePTY::STTY(const char *Command)
  {
    /*
     * Save the new command string to initialize the TTY.
     */

    if (Status == PTY_UNINITIALIZED)
        {
	  delete STTYCommand;
	  STTYCommand = strcpy( new char[ strlen( Command ) + 1 ], Command );
	}
      else Status = BAD_STTY;
  }


int PipePTY::Wait()
  {
    /*
     * Wait for the subprocess to complete.  Of course, it only works
     * if the ChildID is not the same as this process ID.
     */

    if (ChildID == getpid())
        return( OK );

    if (Status != PTY_CLOSED)
        {
	  if (Status == OK)
	      Status = PTY_CANNOT_WAIT;
	  return -1;
	}
    int StatusLoc;
    if (waitpid( ChildID, &StatusLoc, 0 ) != ChildID)
        {
	  Status = SYSTEM_ERROR;
	  SystemErrno = errno;
	}
    return( OK );
  }
